from pyvis.network import Network
import matplotlib.pyplot as plt
import numpy as np



def credal_ranking(weights, criteria_name, file_location):
    avg_weights = np.mean(weights, axis=0)
    index = np.argsort(-avg_weights)

    assert len(criteria_name) == weights.shape[1], "Invalid number of criteria"

    sample_no, c_no = weights.shape
    for i in range(c_no):
        criteria_name[i] = criteria_name[i] + ' - ' + str(round(avg_weights[i],3))

    probs=np.empty((c_no, c_no))
    for i in range(c_no):
        for j in range(i, c_no):
            probs[i,j] = round((weights[:,i] >= weights[:,j]).sum() / sample_no,2)
            probs[j,i] = 1 - probs[i,j]

    ## Visualization using pyvis
    net= Network(notebook=False, layout=None, height='800px', width='600px', directed=True)
    for i in range(c_no):
        net.add_node(str(index[i]), size=max(avg_weights[index[i]]*100,10), 
            title=criteria_name[index[i]], label=criteria_name[index[i]], x=0, y=i*200)

    for i in range(c_no-1):
        net.add_edge(str(index[i]), str(index[i+1]), label=str(probs[index[i],index[i+1]]))
        for j in range(i+2, c_no):
            if probs[index[i], index[j]] < 1 and probs[index[i], index[j]] > 0.5:
                net.add_edge(str(index[i]), str(index[j]), label=str(probs[index[i],index[j]]))

    net.toggle_physics(False)
    net.set_edge_smooth("curvedCW")
    #net.show_buttons(filter_=[])
    #net.prep_notebook()

    net.show(file_location)


def weight_distribution(weights, criteria_name, cols=5, row_based=True):
    c_no = len(criteria_name)

    rows =  int(np.ceil(c_no / cols)) # (c_no // cols) + (c_no % cols > 0)

    if not row_based:
        rows, cols = cols, rows 
        fig, axs = plt.subplots(rows,cols, sharey=True, sharex=True, figsize = (3,15))
    else:
        fig, axs = plt.subplots(rows,cols, sharey=True, sharex=True, figsize = (15,3))

    count = 0
    for i in range(rows):
        if count > c_no: break
        for j in range(cols):
            #axs[i,j].hist(weights[:,i], bins = 50)
            #axs[i,j].set_title(criteria_name[i])
            plt.subplot(rows,cols, count+1)
            plt.hist(weights[:,count], bins=50)
            plt.title(criteria_name[count])
            count += 1

    plt.show()




import sys
import os
sys.path.append(os.path.abspath("src"))

import numpy as np
from BayesMCDM import WeightAnalyzer

if __name__ == "__main__":
    prf =  np.array([
        [ 7, 6, 4, 9, 5, 8, 1, 3],
        [ 9, 8, 2, 5, 4, 5, 1, 3],
        [ 8, 8, 5, 9, 5, 5, 1, 2],
        [ 8, 9, 2, 8, 1, 8, 2, 2],
        [ 8, 6, 1, 9, 6, 7, 4, 4],
        [ 9, 8, 1, 9, 7, 5, 5, 6],
    ])
    weights = prf / prf.sum(axis=1, keepdims=True)

    dmNo, cNo = prf.shape
    altNo = 50
    x = np.random.rand(altNo // 2, cNo)
    altMat = np.concatenate([x*10,x])

    opt = {'CriteriaDependence': False, 'Sigma': np.eye(cNo) }

    weight_aggregation = WeightAnalyzer.StandardWeightAnalyzer(weights=weights, opt=opt, alternatives=altMat) #, dm_cluster_number=2)
    weight_aggregation.sampling()
    print('Ok')



import numpy as np
from Models.BWM.StandardBWM import StandardBWM
from Models.BWM.GaussianBWM import GaussianBWM
from Models.BWM.TriangularBWM import TriangularBWM
from Models.BWM.IntervalBWM import IntervalBWM

import warnings
warnings.filterwarnings("ignore", category=FutureWarning)

import logging
logging.getLogger("httpstan").setLevel(logging.ERROR)


num_samples = 200
num_chains = 2

def get_example_data():
    a_b = np.array([
        [3, 4, 6, 1, 5, 2, 9, 7],
        [1, 2, 8, 4, 5, 3, 9, 6],
        [2, 2, 3, 1, 5, 5, 9, 8],
        [2, 1, 8, 2, 9, 3, 8, 8],
        [2, 4, 9, 1, 4, 3, 5, 5],
        [1, 2, 9, 1, 3, 5, 5, 4]
    ])
    a_w = np.array([
        [7, 6, 4, 9, 5, 8, 1, 3],
        [9, 8, 2, 5, 4, 5, 1, 3],
        [8, 8, 5, 9, 5, 5, 1, 2],
        [8, 9, 2, 8, 1, 8, 2, 2],
        [8, 6, 1, 9, 6, 7, 4, 4],
        [9, 8, 1, 9, 7, 5, 5, 6]
    ])
    dmNo, cNo = a_w.shape
    altNo = 50
    x = np.random.rand(altNo // 2, cNo)
    altMat = np.concatenate([x, x * 1])
    return a_b, a_w, dmNo, cNo, altMat

def run_standard_bwm(a_b, a_w, cNo, altMat):
    print("\n--- StandardBWM ---")
    for crit_dep in [False, True]:
        for dm_cluster in [-1,2]:
            opt = {'CriteriaDependence': crit_dep, 'Sigma': np.eye(cNo)}
            print(f"CriteriaDependence={crit_dep}")
            bwm = StandardBWM(AB=a_b*100, AW=a_w*100, opt=opt, alternatives=altMat, dm_cluster_number=dm_cluster, 
                              num_samples=num_samples, num_chain=num_chains)
            bwm.sampling()
    print("StandardBWM run complete.")

def run_gaussian_bwm(a_b, a_w, cNo, altMat):
    print("\n--- GaussianBWM ---")
    for crit_dep in [False, True]:
        for dm_cluster in [-1, 2]:
            opt = {'CriteriaDependence': crit_dep, 'Sigma': np.eye(cNo)}
            print(f"CriteriaDependence={crit_dep}")
            bwm = GaussianBWM(
                AB_md=a_b, AB_sigma=0.01*np.ones(a_b.shape),
                AW_md=a_w, AW_sigma=0.01*np.ones(a_w.shape),
                opt=opt, alternatives=altMat, dm_cluster_number=dm_cluster, 
                num_samples=num_samples, num_chain=num_chains
            )
            bwm.sampling()
        print("GaussianBWM run complete.")

def run_triangular_bwm(a_b, a_w, cNo, altMat):
    print("\n--- TriangularBWM ---")
    for crit_dep in [True, False]:
        for dm_cluster in [-1, 2]:
            opt = {'CriteriaIndependence': crit_dep}
            print(f"CriteriaIndependence={crit_dep}")
            bwm = TriangularBWM(AB_md=a_b, AW_md=a_w, opt=opt,  dm_cluster_number=dm_cluster,
                                num_samples=num_samples, num_chain=num_chains)
            bwm.sampling()
    print("TriangularBWM run complete.")

def run_interval_bwm(a_b, a_w, cNo, altMat):
    print("\n--- IntervalBWM ---")
    for crit_dep in [True, False]:
        for dm_cluster in [-1, 2]:
            opt = {'CriteriaDependence': crit_dep}
            print(f"CriteriaDependence={crit_dep}")
            bwm = IntervalBWM(
                AB_l=a_b, AB_h=a_b, AW_l=a_w, AW_h=a_w, opt=opt, 
                dm_cluster_number=dm_cluster,
                num_samples=num_samples, num_chain=num_chains
            )
            bwm.sampling()
        print("IntervalBWM run complete.")

if __name__ == "__main__":
    a_b, a_w, dmNo, cNo, altMat = get_example_data()
    run_standard_bwm(a_b, a_w, cNo, altMat)
    run_gaussian_bwm(a_b, a_w, cNo, altMat) # dependece model for clustering does not work yet
    run_triangular_bwm(a_b, a_w, cNo, altMat)
    run_interval_bwm(a_b, a_w, cNo, altMat)
    print("\nAll BWM model runs completed.")


from ..SWING.StandardSWING import StandardSWING
from ..MCDMProblem import MCDMProblem
import numpy as np

class TriangularSWING(StandardSWING, MCDMProblem):
    _basic_model = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            array[DmNo] vector[CNo] Prf_md;
            vector<lower=0,upper=1>[CNo] e;
            real<lower=0> gamma_param;
        } 

        parameters { 
            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            array[DmNo]vector<lower=0, upper=1>[CNo] Prf_trnf;            
        }

        transformed parameters {
            array[DmNo] vector<lower=0>[CNo] Prf; 
            array[DmNo] vector<lower=0>[CNo] W; 

            
            for(i in 1:DmNo) {
                Prf[i] = Prf_trnf[i] + Prf_md[i];
                
                W[i] = Prf[i] ./ sum(Prf[i]);       
            }
        } 

        model {
            kappaStar ~ gamma(gamma_param,gamma_param);
            wStar ~ dirichlet(0.01*e);        
            
            for (i in 1:DmNo) {
                W[i] ~ dirichlet(kappaStar*wStar);

                target += log1m(abs( Prf[i] - Prf_md[i] ));
            }
        } 
    """

    _basic_model_clustering = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            array[DmNo] vector[CNo] Prf_md;
            vector<lower=0,upper=1>[CNo] e;
            int<lower=2> DmC;  
            real<lower=0> gamma_param;
        } 

        parameters { 
            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            array[DmNo] vector<lower=0, upper=1>[CNo] Prf_trnf;

            array[DmC] simplex[CNo] wc;
            array[DmC] real<lower=0> ksi; 
            array[DmNo] simplex[DmC] theta;

        } 

        transformed parameters {
            array[DmNo] vector<lower=0>[CNo] Prf; 
            array[DmNo] vector<lower=0>[CNo] W; 

            
            for(i in 1:DmNo) {
                Prf[i] = Prf_trnf[i] + Prf_md[i];
                W[i] = Prf[i] ./ sum(Prf[i]);       
            }
        } 

        model {
            
            for(i in 1:DmC){
                wc[i] ~ dirichlet(0.01*e);
                ksi[i] ~ gamma(gamma_param,gamma_param);
            }

            for (i in 1:DmNo) {
                array[DmC] real contribution;
                for(j in 1:DmC)
                     contribution[j] = log(theta[i,j]) + dirichlet_lpdf( W[i] | ksi[j]*wc[j]);
                target += log_sum_exp(contribution);

                target += log1m(abs( Prf[i] - Prf_md[i] ));
                
            }
        }     
    """

    _basicModelSorting = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            vector[CNo] AB_md[DmNo];
            vector[CNo] AW_md[DmNo];  
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            int<lower=2> AltC;
            matrix[AltNo, CNo] Alt;
        } 

        parameters { 
            simplex[CNo] W[DmNo];
            real<lower=0> kappa[DmNo]; 

            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            vector[AltC] altMu;

            vector<lower=0, upper=1>[CNo] AW_trnf[DmNo];
            vector<lower=0, upper=1>[CNo] AB_trnf[DmNo];
        } 

        transformed parameters {
            vector<lower=0>[CNo] AB[DmNo]; 
            vector<lower=0>[CNo] AW[DmNo];
            simplex[CNo] AB_normalized[DmNo];
            simplex[CNo] AW_normalized[DmNo];

            for(i in 1:DmNo) {
                AB[i] = AB_trnf[i] + AB_md[i];
                AW[i] = AW_trnf[i] + AW_md[i];

                AB_normalized[i] = (e ./ AB[i]) ./ sum(e ./ AB[i]); 
                AW_normalized[i] = AW[i] ./ sum(AW[i]);       
            }

            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real<upper=0> soft_z; // log unnormalized clusters
            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);
        } 

        model {
            kappaStar ~ gamma(.01,.01);
            wStar ~ dirichlet(0.01*e);
            kappa ~ gamma(.01,.01);
                            
            for (i in 1:DmNo) {
                W[i] ~ dirichlet(kappaStar*wStar);

                target += log1m(fabs( AB[i] - AB_md[i] ));
                target += log1m(fabs( AW[i] - AW_md[i] ));

                AW_normalized[i] ~ dirichlet(kappa[i]*W[i]);
                AB_normalized[i] ~ dirichlet(kappa[i]*W[i]);   
            }
            
            for (n in 1:AltNo)
                target += log_sum_exp(soft_z[n]);
        } 
        """
    
    _correlatedModel = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            vector[CNo] AB_md[DmNo];
            vector[CNo] AW_md[DmNo]; 
            vector<lower=0,upper=1>[CNo] e;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        } 

        parameters { 
            vector[CNo] W_eta[DmNo]; 
            real<lower=0> kappa[DmNo]; 

            vector[CNo] wStar_eta;
            real<lower=0> kappaStar;
            
            vector<lower=0, upper=1>[CNo] AW_trnf[DmNo];
            vector<lower=0, upper=1>[CNo] AB_trnf[DmNo];
        } 

        transformed parameters {
            vector<lower=0>[CNo] AB[DmNo]; 
            vector<lower=0>[CNo] AW[DmNo];
            simplex[CNo] AB_normalized[DmNo];
            simplex[CNo] AW_normalized[DmNo];

            for(i in 1:DmNo) {
                AB[i] = AB_trnf[i] + AB_md[i];
                AW[i] = AW_trnf[i] + AW_md[i];

                AB_normalized[i] = (e ./ AB[i]) ./ sum(e ./ AB[i]); 
                AW_normalized[i] = AW[i] ./ sum(AW[i]);         
            }
            
            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);
            
            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);
            
        } 

        model {
            wStar_eta ~ multi_normal(mu, Sigma);
            kappa ~ gamma(.01,.01);
                            
            for (i in 1:DmNo) {
                W_eta[i] ~ multi_normal(wStar_eta, Sigma);

                target += log1m(fabs( AB[i] - AB_md[i] ));
                target += log1m(fabs( AW[i] - AW_md[i] ));

                AW_normalized[i] ~ dirichlet(kappa[i]*W[i]);
                AB_normalized[i] ~ dirichlet(kappa[i]*W[i]);    
            }
        }
    """

    _correlatedModelClustering = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            vector[CNo] AB_md[DmNo];
            vector[CNo] AW_md[DmNo];   
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        } 

        parameters { 
            vector[CNo] W_eta[DmNo];
            real<lower=0> kappa[DmNo]; 
            
            simplex[DmC] theta[DmNo];  
            vector[CNo] wc_eta[DmC];
            real<lower=0> ksi[DmC];

            real<lower=0> kappaStar;
 
            vector<lower=0, upper=1>[CNo] AW_trnf[DmNo];
            vector<lower=0, upper=1>[CNo] AB_trnf[DmNo];
        } 

        transformed parameters {
            vector<lower=0>[CNo] AB[DmNo]; 
            vector<lower=0>[CNo] AW[DmNo];
            simplex[CNo] AB_normalized[DmNo];
            simplex[CNo] AW_normalized[DmNo];

            for(i in 1:DmNo) {
                AB[i] = AB_trnf[i] + AB_md[i];
                AW[i] = AW_trnf[i] + AW_md[i];

                AB_normalized[i] = (e ./ AB[i]) ./ sum(e ./ AB[i]); 
                AW_normalized[i] = AW[i] ./ sum(AW[i]);        
            }

            
            simplex[CNo] W[DmNo];
            simplex[CNo] wc[DmC];

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);

            for(i in 1:DmC)
                wc[i] = softmax(wc_eta[i]);
        } 

        model {
            kappa ~ gamma(.01,.01);
            ksi ~ gamma(0.01, 0.01);
            for(i in 1:DmC)
                wc_eta[i] ~ multi_normal(mu, Sigma);
                           
            for (i in 1:DmNo) {
                real contribution[DmC];
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + multi_normal_lpdf( W_eta[i] | wc_eta[j], Sigma);
                target += log_sum_exp(contribution);   

                target += log1m(fabs( AB[i] - AB_md[i] ));
                target += log1m(fabs( AW[i] - AW_md[i] ));

                AW_normalized[i] ~ dirichlet(kappa[i]*W[i]);
                AB_normalized[i] ~ dirichlet(kappa[i]*W[i]);    
            }
        }
    """

    _correlatedModelSorting = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            vector[CNo] AB_md[DmNo];
            vector[CNo] AW_md[DmNo];  
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            matrix[AltNo, CNo] Alt;
            int<lower=2> AltC;
            vector<lower=0,upper=1>[AltC] eAlt;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        } 

        parameters { 
            vector[CNo] W_eta[DmNo]; 
            real<lower=0> kappa[DmNo]; 
 
            vector[CNo] wStar_eta;
            real<lower=0> kappaStar; 

            vector<lower=0, upper=1>[CNo] AW_trnf[DmNo];
            vector<lower=0, upper=1>[CNo] AB_trnf[DmNo];
       
            vector[AltC] altMu;
        } 

        transformed parameters {
            vector<lower=0>[CNo] AB[DmNo]; 
            vector<lower=0>[CNo] AW[DmNo];
            simplex[CNo] AB_normalized[DmNo];
            simplex[CNo] AW_normalized[DmNo];

            for(i in 1:DmNo) {
                AB[i] = AB_trnf[i] + AB_md[i];
                AW[i] = AW_trnf[i] + AW_md[i];

                AB_normalized[i] = (e ./ AB[i]) ./ sum(e ./ AB[i]); 
                AW_normalized[i] = AW[i] ./ sum(AW[i]);      
            }
            
            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);
            
            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);

            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real<upper=0> soft_z; // log unnormalized clusters
            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);
        } 

        model {
            wStar_eta ~ multi_normal(mu, Sigma);
            kappa ~ gamma(.01,.01);
                            
            for (i in 1:DmNo) {
                W_eta[i] ~ multi_normal(wStar_eta, Sigma);

                target += log1m(fabs( AB[i] - AB_md[i] ));
                target += log1m(fabs( AW[i] - AW_md[i] ));
                
                AW_normalized[i] ~ dirichlet(kappa[i]*W[i]);
                AB_normalized[i] ~ dirichlet(kappa[i]*W[i]);    
            }

            for (n in 1:AltNo)
               target += log_sum_exp(soft_z[n]);

        }
    """ 
    
    def __init__(self, Prf_md, alternatives = None, dm_cluster_number=-1, alt_sort_number=-1, num_chain=3, num_samples=1000, opt={}):
        self.Prf_MD = np.array(Prf_md)

        MCDMProblem.__init__(self, alternatives, dm_cluster_number, alt_sort_number, num_chain, num_samples, opt)


    @property
    def input_data(self):

        data = self._get_common_data()
        data['Prf_md'] = self.Prf_MD
        data['e'] =  np.ones(self.criteria_no)

        return data

    
    @property 
    def original_model(self):
        return self.__originalModel

    @property
    def dm_no(self):
        return self.Prf_MD.shape[0]

    @property
    def criteria_no(self):
        return self.Prf_MD.shape[1]

    def _check_input_data(self):

        assert self.Prf_MD.shape[0] >=1, "No input"
        assert self.Prf_MD.shape[1] >=2, "The number of criteria must be more than 2!"

        return True




from .StandardSWING import StandardSWING
from .GaussianSWING import GaussianSWING
from .IntervalSWING import IntervalSWING
from .TriangularSWING import TriangularSWING

__all__ = ["StandardSWING", "GaussianSWING", "IntervalSWING", "TriangularSWING"]



from ..MCDMProblem import MCDMProblem
import numpy as np

class StandardWeightAnalyzer(MCDMProblem):
    _basic_model = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] W;
            vector<lower=0,upper=1>[CNo] e;
            real<lower=0> gamma_param;
        }

        parameters {
            simplex[CNo] wStar;
            real<lower=0> kappaStar;
        }

        transformed parameters {
    
        } 

        model {
            wStar ~ dirichlet(0.01*e);
            kappaStar ~ gamma(gamma_param,gamma_param);

            for (i in 1:DmNo){
                W[i] ~ dirichlet(kappaStar*wStar);
            }
        }
    """

    _basic_model_clustering = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] W;
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;
            real<lower=0> gamma_param;
        }

        parameters {
            array[DmC] simplex[CNo] wc;
            array[DmC] real<lower=0> ksi;
            array[DmNo] simplex[DmC] theta;
        }

        transformed parameters {
        } 

        model {
            for(i in 1:DmC){
                wc[i] ~ dirichlet(0.001*e);
                ksi[i] ~ gamma(gamma_param, gamma_param);
            }

            for (i in 1:DmNo) {
                array[DmC] real contribution;
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + dirichlet_lpdf( W[i] | ksi[j]*wc[j]);
                target += log_sum_exp(contribution);

            }
        }
    """

    _basicModelSorting = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            vector[CNo] Prf[DmNo];
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            int<lower=2> AltC;
            matrix[AltNo, CNo] Alt;
        }

        parameters {
            simplex[CNo] W[DmNo];
            real<lower=0> kappa[DmNo]; 

            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            vector[AltC] altMu;
        }

        transformed parameters {
            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real soft_z;
            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);

            simplex[CNo] Prf_normalized[DmNo];
            for(i in 1:DmNo) {
                Prf_normalized[i] = Prf[i] ./ sum(Prf[i]);       
            }

        }

        model {
            // Prior
            wStar ~ dirichlet(0.01*e);
            kappaStar ~ gamma(.001,.01);
            W ~ dirichlet(kappaStar*wStar);
            kappa ~ gamma(.01,.01);
            altMu ~ normal(0,5);

            for (i in 1:DmNo){
            
                Prf_normalized[i] ~ dirichlet(kappa[i]*W[i]);  
            }

            for (n in 1:AltNo)
                target += log_sum_exp(soft_z[n]);
        }
    """

    _correlated_model = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] W;
            vector<lower=0,upper=1>[CNo] e;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
            
            real<lower=0> gamma_param; // parameter for gamma distribution
            real<lower=0> sigma_coef; // parameter for sigma
        }

        parameters {
            vector[CNo] wStar_eta;
            real<lower=0> kappaStar;           
        }

        transformed parameters {
            array[DmNo] vector[CNo] W_eta;
            for(i in 1:DmNo){
                vector[CNo] logW = log(W[i]);
                real mean_logW = mean(logW);
                W_eta[i] = logW - mean_logW;  // inverse softmax with zero-mean
            }

            simplex[CNo] wStar;
            wStar = softmax(wStar_eta);
        }

        model {
            wStar_eta ~ multi_normal(mu, Sigma);

            for (i in 1:DmNo){
                W_eta[i] ~ multi_normal(wStar_eta, sigma_coef*Sigma);
            }
        }
    """

    _correlated_model_clustering = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] W;
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;

            real<lower=0> gamma_param; // parameter for gamma distribution
            real<lower=0> sigma_coef; // parameter for sigma
        }

        parameters {
            array[DmC] vector[CNo] wc_eta;
            array[DmNo] simplex[DmC] theta;
        }

        transformed parameters {
            array[DmNo] vector[CNo] W_eta;
            for(i in 1:DmNo){
                vector[CNo] logW = log(W[i]);
                real mean_logW = mean(logW);
                W_eta[i] = logW - mean_logW;  // inverse softmax with zero-mean
            }

            array[DmC] simplex[CNo] wc;
            for(i in 1:DmC)
                wc[i] = softmax(wc_eta[i]);
        }

        model {
            for (d in 1:DmC)
                wc_eta[d] ~ multi_normal(mu, Sigma);

            for (i in 1:DmNo){
                array[DmC] real contribution;
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + multi_normal_lpdf( W_eta[i] | wc_eta[j], Sigma*sigma_coef);
                target += log_sum_exp(contribution);
            }
        }
    """

    _correlatedModelSorting = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] Prf;
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            matrix[AltNo, CNo] Alt;
            int<lower=2> AltC;
            vector<lower=0,upper=1>[AltC] eAlt;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        }

        parameters {
            vector[CNo] W_eta[DmNo];
            real<lower=0> kappa[DmNo]; 

            vector[CNo] wStar_eta;
            real<lower=0> kappaStar;

            vector[AltC] altMu;
        }

        transformed parameters {
            simplex[CNo] Prf_normalized[DmNo];

            for(i in 1:DmNo)
                Prf_normalized[i] = Prf[i] ./ sum(Prf[i]);       

            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);


            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real<upper=0> soft_z; // log unnormalized clusters

            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);

        }

        model {
            wStar_eta ~ multi_normal(mu, Sigma);
            kappaStar ~ gamma(0.001,0.001);
            kappa ~ gamma(0.001, 0.001);

            for (i in 1:DmNo){
                W_eta[i] ~ multi_normal(wStar_eta, Sigma);

                Prf_normalized[i] ~ dirichlet(kappa[i]*W[i]);  
            }

            for (n in 1:AltNo)
                target += log_sum_exp(soft_z[n]);

        }
    """

    def __init__(self, weights, alternatives = None, dm_cluster_number=-1, alt_sort_number=-1, num_chain=3, num_samples=1000, opt={}):
        array_weight = np.array(weights)
        self.weights = weights = array_weight / array_weight.sum(axis=1, keepdims=True)

        super().__init__(alternatives, dm_cluster_number, alt_sort_number, num_chain, num_samples, opt)


    @property
    def input_data(self):
        data = self._get_common_data()
        data['W'] = self.weights
        data['e'] =  np.ones(self.criteria_no)

        return data

    @property
    def original_model(self):
        return self.__originalModel

    @property
    def dm_no(self):
        return self.weights.shape[0]

    @property
    def criteria_no(self):
        return self.weights.shape[1]

    def _check_input_data(self):
        assert self.weights.shape[0] >= 1, "No input"
        assert self.weights.shape[1] >= 2, "The number of criteria must be more than 2!"

        return True





from .StandardPointAllocation import StandardPointAllocation
from .GaussianPointAllocation import GaussianPointAllocation
from .IntervalPointAllocation import IntervalPointAllocation
from .TriangularPointAllocation import TriangularPointAllocation

__all__ = ["StandardPointAllocation", "GaussianPointAllocation", "IntervalPointAllocation", "TriangularPointAllocation"]



from .StandardAHP import StandardAHP
from .GaussianAHP import GaussianAHP
from .IntervalAHP import IntervalAHP
from .TriangularAHP import TriangularAHP

__all__ = ["StandardAHP", "GaussianAHP", "IntervalAHP", "TriangularAHP"]


from ..SWING.TriangularSWING import TriangularSWING

class TriangularPointAllocation(TriangularSWING):
    """
    This class is a direct alias of TriangularSWING for Point Allocation method.
    """
    pass



from ..MCDMProblem import MCDMProblem 
import numpy as np

class StandardAHP(MCDMProblem):

    _basic_model = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] matrix[CNo,CNo] PCM;
            vector<lower=0,upper=1>[CNo] e;
            real<lower=0> gamma_param;
         }

         parameters {
            array[DmNo] simplex[CNo] W;
            array[DmNo] real<lower=0> kappa;

            simplex[CNo] wStar;
            real<lower=0> kappaStar;
         }

         transformed parameters {
            array[DmNo] matrix[CNo,CNo] PCM_normalized;

            for(i in 1:DmNo) 
                for(j in 1:CNo)
                    PCM_normalized[i][,j] = col(PCM[i],j) ./  sum(col(PCM[i],j));
         } 

         model {
            kappaStar ~ gamma(gamma_param,gamma_param);
            wStar ~ dirichlet(0.01*e);

            for (i in 1:DmNo){
                W[i] ~ dirichlet(kappaStar*wStar);
                kappa[i] ~ gamma(gamma_param,gamma_param);

                for(j in 1:CNo){
                    PCM_normalized[i][,j] ~ dirichlet(kappa[i]*W[i]);
                }   
            }
         }
    """

    _basic_model_clustering = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] matrix[CNo,CNo] PCM;
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;
            real<lower=0> gamma_param;
        }

        parameters {
            array[DmNo] simplex[CNo] W;
            array[DmNo] real<lower=0> kappa;

            array[DmC] simplex[CNo] wc;
            array[DmC] real<lower=0> ksi;
            array[DmNo] simplex[DmC] theta;
        }

        transformed parameters {
            array[DmNo] matrix[CNo,CNo] PCM_normalized;

            for(i in 1:DmNo) 
                for(j in 1:CNo)
                    PCM_normalized[i][,j] = col(PCM[i],j) ./  sum(col(PCM[i],j));
        } 

        model {
            for(i in 1:DmC){
                wc[i] ~ dirichlet(0.01*e);
                ksi[i] ~ gamma(gamma_param,gamma_param);
            }
            
            for (i in 1:DmNo) {
                array[DmC] real contribution;
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + dirichlet_lpdf( W[i] | ksi[j]*wc[j]);
                target += log_sum_exp(contribution);

                kappa[i] ~ gamma(gamma_param,gamma_param);
                for(j in 1:CNo)
                    PCM_normalized[i][,j] ~ dirichlet(kappa[i]*W[i]); 
            }
        }
    """

    _correlated_model = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] matrix[CNo,CNo] PCM;
            vector<lower=0,upper=1>[CNo] e;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;

            real<lower=0> gamma_param; // parameter for gamma distribution
            real<lower=0> sigma_coef; // parameter for sigma
        }

        parameters {
            array[DmNo] vector[CNo] W_eta;
            array[DmNo] real<lower=0> kappa;

            vector[CNo] wStar_eta;
        }

        transformed parameters {
            array[DmNo] matrix[CNo,CNo] PCM_normalized;

            for(i in 1:DmNo) 
                for(j in 1:CNo)
                    PCM_normalized[i][,j] = col(PCM[i],j) ./  sum(col(PCM[i],j));

            array[DmNo] simplex[CNo] W;
            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);
        }

        model {
            wStar_eta ~ multi_normal(mu, Sigma);

            for (i in 1:DmNo){
                W_eta[i] ~ multi_normal(wStar_eta, sigma_coef*Sigma);
                kappa[i] ~ gamma(gamma_param,gamma_param);

                for(j in 1:CNo)
                    PCM_normalized[i][,j] ~ dirichlet(kappa[i]*W[i]); 
            }
        }
    """

    _correlated_model_clustering = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] matrix[CNo,CNo] PCM;
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;

            real<lower=0.0001> gamma_param; // prior for the Gamma distribution
            real<lower=0.0001> sigma_coef; // coefficient for the covariance matrix
        }

        parameters {
            array[DmNo] vector[CNo] W_eta;
            array[DmNo] real<lower=0> kappa;

            array[DmC] vector[CNo] wc_eta;
            array[DmNo] simplex[DmC] theta;
        }

        transformed parameters {
            for(i in 1:DmNo) 
                for(j in 1:CNo)
                    PCM_normalized[i][,j] = col(PCM[i],j) ./  sum(col(PCM[i],j));

            array[DmNo] simplex[CNo] W;
            array[DmC] simplex[CNo] wc;

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);

            for(i in 1:DmC)
                wc[i] = softmax(wc_eta[i]);
        }

        model {
            wStar_eta ~ multi_normal(mu, Sigma);

            for (d in 1:DmC){
                wc_eta[d] ~ multi_normal(mu, Sigma);
            }

            for (i in 1:DmNo){

                array[DmC] real contribution;
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + multi_normal_lpdf( W_eta[i] | wc_eta[j], Sigma*sigma_coef);
                target += log_sum_exp(contribution);

                kappa[i] ~ gamma(gamma_param,gamma_param);
                for(j in 1:CNo)
                    PCM_normalized[i][,j] ~ dirichlet(kappa[i]*W[i]); 
            }
        }
    """

    def __init__(self, PCM, alternatives = None, dm_cluster_number=-1, alt_sort_number=-1, num_chain=3, num_samples=1000, opt={}):
        self.PCM = np.array(PCM)
        
        super().__init__(alternatives, dm_cluster_number, alt_sort_number, num_chain, num_samples, opt)


    @property
    def input_data(self):
        data = self._get_common_data()
        data['PCM'] = self.PCM
        data['e'] =  np.ones(self.criteria_no)

        return data

    @property
    def original_model(self):
        return self.__originalModel

    @property
    def dm_no(self):
        return self.PCM.shape[0]

    @property
    def criteria_no(self):
        return self.PCM[0].shape[1]

    def _check_input_data(self):

        return True

